#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 - 2020, 2024 Pytroll Community

# Author(s):

#   Adam.Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Drawing satellite overpass outlines on maps."""

import logging
import logging.handlers
import os
from tempfile import gettempdir

import matplotlib as mpl
import numpy as np

MPL_BACKEND = mpl.get_backend()

logger = logging.getLogger(__name__)

try:
    import cartopy.crs as ccrs
    import cartopy.feature as cfeature
    BASEMAP_NOT_CARTOPY = False
except ImportError:
    logger.warning("Failed loading Cartopy, will try Basemap instead")
    BASEMAP_NOT_CARTOPY = True

if not BASEMAP_NOT_CARTOPY:
    import cartopy
    cartopy.config["pre_existing_data_dir"] = os.environ.get(
        "CARTOPY_PRE_EXISTING_DATA_DIR", cartopy.config["pre_existing_data_dir"])


class MapperBasemap:
    """A class to generate nice plots with basemap."""

    def __init__(self, **proj_info):
        """Initialize the class object."""
        from mpl_toolkits.basemap import Basemap

        if not proj_info:
            proj_info = {
                "projection": "nsper",
                "lat_0": 58,
                "lon_0": 16,
                "resolution": "l",
                "area_thresh": 1000.
            }

        self.map = Basemap(**proj_info)

        self.map.drawcoastlines()
        self.map.drawcountries()

        self.map.drawmapboundary(fill_color="white")

        self.map.drawmeridians(np.arange(0, 360, 30))
        self.map.drawparallels(np.arange(-90, 90, 30))

    def __enter__(self):
        """Enter the mapper context."""
        return self.map

    def __exit__(self, etype, value, tb):
        """Exit the mapper context."""
        pass


class MapperCartopy:
    """A class to generate nice plots with Cartopy."""

    def __init__(self, **proj_info):
        """Initialize the Cartopy Mapper object."""
        mpl.use(MPL_BACKEND)
        import matplotlib.pyplot as plt

        if not proj_info:
            proj_info = {
                "central_latitude": 58,
                "central_longitude": 16,
                "satellite_height": 35785831,
                "false_easting": 0,
                "false_northing": 0,
                "globe": None
            }

        fig = plt.figure(figsize=(8, 6))

        self._ax = fig.add_subplot(
            1, 1, 1, projection=ccrs.NearsidePerspective(**proj_info))

        self._ax.add_feature(cfeature.OCEAN, zorder=0)
        self._ax.add_feature(cfeature.LAND, zorder=0, edgecolor="black")
        self._ax.add_feature(cfeature.BORDERS, zorder=0)

        self._ax.set_global()
        self._ax.gridlines()

    def plot(self, *args, **kwargs):
        """Make the matplotlib plot."""
        mpl.use(MPL_BACKEND)
        import matplotlib.pyplot as plt

        kwargs["transform"] = ccrs.Geodetic()
        return plt.plot(*args, **kwargs)

    def nightshade(self, utctime, **kwargs):
        """Create a night shade to the map."""
        from trollsched.helper_functions import fill_dark_side

        color = kwargs.get("color", "black")
        alpha = kwargs.get("alpha", 0.4)
        fill_dark_side(self._ax, time=utctime, color=color, alpha=alpha)

    def __call__(self, *args):
        """Call the mapper."""
        return args

    def __enter__(self):
        """Enter the mapper context."""
        return self

    def __exit__(self, etype, value, tb):
        """Exit the mapper context."""
        pass


Mapper = MapperBasemap if BASEMAP_NOT_CARTOPY else MapperCartopy


def save_fig(pass_obj,
             poly=None,
             directory=None,
             overwrite=False,
             labels=None,
             extension=".png",
             outline="-r",
             plot_parameters=None,
             plot_title=None,
             poly_color=None):
    """Save the pass as a figure. Filename is automatically generated."""
    poly = poly or []
    poly_color = poly_color or []
    if not isinstance(poly, (list, tuple)):
        poly = [poly]
    if not isinstance(poly_color, (list, tuple)):
        poly_color = [poly_color]

    mpl.use("Agg")
    import matplotlib.pyplot as plt
    plt.clf()

    logger.debug("Save fig " + str(pass_obj))
    rise = pass_obj.risetime.strftime("%Y%m%d%H%M%S")
    fall = pass_obj.falltime.strftime("%Y%m%d%H%M%S")
    if directory is None:
        directory = os.path.join(gettempdir(), "plots")
    if not os.path.exists(directory):
        logger.debug("Create plot dir " + directory)
        os.makedirs(directory)

    filename = "{rise}_{satname}_{instrument}_{fall}{extension}".format(rise=rise,
                                                                        satname=pass_obj.satellite.name.replace(
                                                                            " ", "_"),
                                                                        instrument=pass_obj.instrument.replace(
                                                                            "/", "-"),
                                                                        fall=fall, extension=extension)
    filepath = os.path.join(directory, filename)
    pass_obj.fig = filepath
    if not overwrite and os.path.exists(filepath):
        return filepath

    logger.debug("Filename = <%s>", filepath)
    plot_parameters = plot_parameters or {}
    with Mapper(**plot_parameters) as mapper:
        mapper.nightshade(pass_obj.uptime, alpha=0.2)
        for i, polygon in enumerate(poly):
            try:
                col = poly_color[i]
            except IndexError:
                col = "-b"
            draw(polygon, mapper, col)
        logger.debug("Draw: outline = <%s>", outline)
        draw(pass_obj.boundary.contour_poly, mapper, outline)

    logger.debug("Title = %s", str(pass_obj))
    if not plot_title:
        plt.title(str(pass_obj))
    else:
        plt.title(plot_title)
    for label in labels or []:
        plt.figtext(*label[0], **label[1])
    logger.debug("Save plot...")
    plt.savefig(filepath)
    logger.debug("Return...")
    return filepath


def show(pass_obj,
         poly=None,
         labels=None,
         other_poly=None,
         proj=None,
         outline="-r"):
    """Show the current pass on screen (matplotlib, basemap)."""
    mpl.use(MPL_BACKEND)
    import matplotlib.pyplot as plt

    proj = proj or {}
    with Mapper(**proj) as mapper:
        mapper.nightshade(pass_obj.uptime, alpha=0.2)
        draw(pass_obj.boundary.contour_poly, mapper, outline)
        if poly is not None:
            draw(poly, mapper, "-b")
        if other_poly is not None:
            draw(other_poly, mapper, "-g")
    plt.title(str(pass_obj))
    for label in (labels or []):
        plt.figtext(*label[0], **label[1])
    plt.show()


def draw(poly, mapper, options, **more_options):
    """Draw the polygon onto the mapper object."""
    lons = np.rad2deg(poly.lon.take(np.arange(len(poly.lon) + 1), mode="wrap"))
    lats = np.rad2deg(poly.lat.take(np.arange(len(poly.lat) + 1), mode="wrap"))
    rx, ry = mapper(lons, lats)
    mapper.plot(rx, ry, options, **more_options)


def main():
    """A main function as a how-to example."""
    from datetime import datetime

    from trollsched.satpass import get_next_passes

    passes = get_next_passes(["noaa 19", "suomi npp"], datetime.now(), 24, (16, 58, 0))

    directory = os.path.join(gettempdir(), "plots")
    for p in passes:
        save_fig(p, directory=directory)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2024 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c22526.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Fixtures for unit testing."""


from datetime import datetime, timedelta

import pytest

from trollsched.satpass import create_pass

# Create a fake TLE file for testing.
TEST_TLEFILE1 = """NOAA-20
1 43013U 17073A   24260.23673431  .00000000  00000+0  20543-3 0 00012
2 43013  98.7102 196.6451 0000981  27.6583 181.3633 14.19588904353826
"""

FAKE_XML_SCHEDULE = """<acquisition-schedule><properties><project>Pytroll</project><type>request</type><station>nrk</station><file-start>2024-12-02-21:54:34</file-start><file-end>2024-12-02-23:54:34</file-end><requested-by>SMHI</requested-by><requested-on>2024-12-02-19:55:05</requested-on></properties><pass satellite="noaa18" start-time="2024-12-02-22:01:31" end-time="2024-12-02-22:16:17" /><pass satellite="npp" start-time="2024-12-02-22:38:18" end-time="2024-12-02-22:49:39" /><pass satellite="noaa20" start-time="2024-12-02-23:01:26" end-time="2024-12-02-23:14:17" /></acquisition-schedule>"""  # noqa

TEST_TLEFILE2 = """NOAA-20
1 43013U 17073A   24336.73680280  .00000000  00000+0  21070-3 0 00017
2 43013  98.7240 271.9468 0002361  91.7021 243.0836 14.19579317364671
NOAA 18
1 28654U 05018A   24336.77887833  .00000603  00000+0  34401-3 0  9991
2 28654  98.8621  52.2610 0013351 310.3593  49.6413 14.13439071  6932
SUOMI NPP
1 37849U 11061A   24336.86623462  .00000353  00000+0  18825-3 0  9992
2 37849  98.7391 272.4265 0002092 106.9131 253.2275 14.19509233678657
"""

@pytest.fixture
def fake_tle_file(tmp_path):
    """Write fake TLE file."""
    file_path = tmp_path / "sometles.txt"
    with open(file_path, "w") as fpt:
        fpt.write(TEST_TLEFILE1)

    return file_path

@pytest.fixture
def fake_long_tle_file(tmp_path):
    """Write fake TLE file."""
    file_path = tmp_path / "some_more_tles.txt"
    with open(file_path, "w") as fpt:
        fpt.write(TEST_TLEFILE2)

    return file_path

@pytest.fixture
def fake_noaa20_viirs_pass_instance(fake_tle_file):
    """Create a fake trollsched.,satpass instance for NOAA-20 VIIRS."""
    starttime = datetime(2024, 9, 17, 1, 25, 52)
    endtime = starttime + timedelta(minutes=15)
    return create_pass("NOAA-20", "viirs", starttime, endtime, str(fake_tle_file))

@pytest.fixture
def fake_xml_schedule_file(tmp_path):
    """Create a fake XML schedule file."""
    file_path = tmp_path / "20241202-195434-aquisition-schedule-request-nrk.xml"
    with open(file_path, "w") as fpt:
        fpt.write(FAKE_XML_SCHEDULE)

    return file_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 - 2024 Pytroll Community
# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Alexander Maul <alexander.maul@dwd.de>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Satellite passes."""

import ftplib
import glob
import hashlib
import logging
import logging.handlers
import operator
import os
import socket
import ssl
import sys
from datetime import datetime, timedelta
from functools import reduce as fctools_reduce
from tempfile import gettempdir, mkstemp
from urllib.parse import urlparse

import numpy as np
from pyorbital import orbital, tlefile
from pyresample.boundary import AreaDefBoundary

from trollsched import (
    JPSS_TLE_NAMES,
    MERSI2_PLATFORM_NAMES,
    MERSI3_PLATFORM_NAMES,
    MERSI_PLATFORM_NAMES,
    MIN_PASS,
    NUMBER_OF_FOVS,
    SATELLITE_MEOS_TRANSLATION,
    VIIRS_PLATFORM_NAMES,
)
from trollsched.boundary import SwathBoundary
from trollsched.pass_scheduling_utils import Satellite

logger = logging.getLogger(__name__)


def create_pass(satname, instrument, starttime, endtime, tle_filename=None):
    """Create a satellite pass given a start and an endtime."""
    tle = tlefile.Tle(satname, tle_file=tle_filename)
    return Pass(satname, starttime, endtime, instrument=instrument, tle1=tle.line1, tle2=tle.line2)

class SimplePass:
    """A pass: satellite, risetime, falltime, (orbital)."""

    buffer = timedelta(minutes=2)

    def __init__(self, satellite, risetime, falltime):
        """Initialize the simple pass."""
        if not hasattr(satellite, "name"):
            self.satellite = Satellite(satellite, 0, 0)
        else:
            self.satellite = satellite
        self.risetime = risetime
        self.falltime = falltime
        self.score = {}
        self.subsattrack = {"start": None, "end": None}
        self.rec = False
        self.fig = None

    def __hash__(self):
        """Hash the pass."""
        return super.__hash__(self)

    def overlaps(self, other, delay=None):
        """Check if two passes overlap in time."""
        if delay is None:
            delay = timedelta(seconds=0)
        return ((self.risetime < other.falltime + delay) and (self.falltime + delay > other.risetime))

    def __lt__(self, other):
        """Check if this pass starts earlier than the other one."""
        return self.uptime < other.uptime

    def __gt__(self, other):
        """Check if this pass startss later than the other one."""
        return self.uptime > other.uptime

    def __cmp__(self, other):
        """Compare two passes."""
        if self.uptime < other.uptime:
            return -1
        if self.uptime > other.uptime:
            return 1
        else:
            return 0

    def __eq__(self, other):
        """Determine if two satellite passes are the same."""
        # Two passes, maybe observed from two distinct stations, are compared by
        # a) satellite name and orbit number,
        # or if the later is not available
        # b) the time difference between rise- and fall-times.
        if other is not None and isinstance(self, Pass) and isinstance(
                other, Pass):
            return (self.satellite.name == other.satellite.name and
                    self.orb.get_orbit_number(self.risetime) == other.orb.get_orbit_number(other.risetime))
        timedelta(seconds=1)
        return (other is not None and
                self.satellite.name == other.satellite.name and
                self.overlaps(other))

    def __str__(self):
        """Give a string version of the pass."""
        return (self.satellite.name + " " + self.risetime.isoformat() + " " +
                self.falltime.isoformat())

    def __repr__(self):
        """Represent the pass."""
        return str(self)

    def duration(self):
        """Get the duration of an overpass."""
        return self.falltime - self.risetime

    def seconds(self):
        """Get the duration of an overpass."""
        duration = self.duration()
        return (duration.days * 24 * 60 * 60 + duration.seconds +
                duration.microseconds * 1e-6)


class Pass(SimplePass):
    """A pass: satellite, risetime, falltime, (orbital)."""

    def __init__(self, satellite, risetime, falltime, **kwargs):
        """Initialize the pass."""
        SimplePass.__init__(self, satellite, risetime, falltime)

        logger.debug("kwargs: %s", str(kwargs))
        orb = kwargs.get("orb", None)
        uptime = kwargs.get("uptime", None)
        instrument = kwargs.get("instrument", None)
        tle1 = kwargs.get("tle1", None)
        tle2 = kwargs.get("tle2", None)
        logger.debug("instrument: %s", str(instrument))

        if isinstance(instrument, (list, set)):
            if "avhrr" in instrument:
                logger.warning("Instrument is a sequence! Assume avhrr...")
                instrument = "avhrr"
            elif "viirs" in instrument:
                logger.warning("Instrument is a sequence! Assume viirs...")
                instrument = "viirs"
            elif "modis" in instrument:
                logger.warning("Instrument is a sequence! Assume modis...")
                instrument = "modis"
            elif "mersi" in instrument:
                logger.warning("Instrument is a sequence! Assume mersi...")
                instrument = "mersi"
            elif "mersi-2" in instrument:
                logger.warning("Instrument is a sequence! Assume mersi-2...")
                instrument = "mersi-2"
            elif "mersi-3" in instrument:
                logger.warning("Instrument is a sequence! Assume mersi-3...")
                instrument = "mersi-3"
            else:
                raise TypeError("Instrument is a sequence! Don't know which one to choose!")

        default = NUMBER_OF_FOVS.get(instrument, 2048)
        self.number_of_fovs = kwargs.get("number_of_fovs", default)
        # The frequency shouldn't actualy depend on the number of FOVS along a scanline should it!?
        # frequency = kwargs.get('frequency', int(self.number_of_fovs / 4))
        frequency = kwargs.get("frequency", 300)

        self.station = None
        self.max_elev = None
        self.uptime = uptime or (risetime + (falltime - risetime) / 2)
        self.instrument = instrument
        self.frequency = frequency
        if orb:
            self.orb = orb
        else:
            try:
                self.orb = orbital.Orbital(satellite, line1=tle1, line2=tle2)
            except KeyError as err:
                logger.debug("Failed in PyOrbital: %s", str(err))
                self.orb = orbital.Orbital(
                    JPSS_TLE_NAMES.get(satellite, satellite),
                    line1=tle1,
                    line2=tle2)
                logger.info("Using satellite name %s instead",
                            str(JPSS_TLE_NAMES.get(satellite, satellite)))

        self._boundary = None

    @property
    def boundary(self):
        """Get the boundary of the swath."""
        if not self._boundary:
            self._boundary = SwathBoundary(self, frequency=self.frequency)
        return self._boundary

    @boundary.setter
    def boundary(self, value):
        self._boundary = SwathBoundary(self, frequency=self.frequency)

    def pass_direction(self):
        """Get the direction of the pass in (ascending, descending)."""
        start_lat = self.orb.get_lonlatalt(self.risetime)[1]
        end_lat = self.orb.get_lonlatalt(self.falltime)[1]

        if start_lat > end_lat:
            return "descending"
        else:
            return "ascending"

    def slsearch(self, sublat):
        """Find sublatitude."""

        def nadirlat(minutes):
            return self.orb.get_lonlatalt(self.risetime + timedelta(
                minutes=np.float64(minutes)))[1] - sublat

        def get_root(fun, start, end):
            p = np.polyfit(
                [start, (start + end) / 2.0, end],
                [fun(start), fun((start + end) / 2),
                 fun(end)], 2)
            for root in np.roots(p):
                if root <= end and root >= start:
                    return root

        arr = np.array([nadirlat(m) for m in range(15)])
        a = np.where(np.diff(np.sign(arr)))[0]
        for guess in a:
            sublat_mins = get_root(nadirlat, guess, guess + 1)
            return self.risetime + timedelta(minutes=sublat_mins)

    def area_coverage(self, area_of_interest):
        """Get the ratio of coverage (between 0 and 1) of the pass with the area of interest."""
        try:
            area_boundary = area_of_interest.poly
        except AttributeError:
            area_boundary = AreaDefBoundary(area_of_interest, frequency=100)
            area_boundary = area_boundary.contour_poly

        inter = self.boundary.contour_poly.intersection(area_boundary)

        if inter is None:
            return 0
        return inter.area() / area_boundary.area()

    def generate_metno_xml(self, coords, root):
        """Generate a metno xml schedule."""
        import xml.etree.ElementTree as ET  # noqa because defusedxml has no SubElement

        asimuth_at_max_elevation, max_elevation = self.orb.get_observer_look(self.uptime, *coords)
        pass_direction = self.pass_direction().capitalize()[:1]
        # anl = self.orb.get_lonlatalt(self.orb.get_last_an_time(self.risetime))[0] % 360
        asimuth_at_aos, aos_elevation = self.orb.get_observer_look(self.risetime, *coords)
        orbit = self.orb.get_orbit_number(self.risetime)
        # aos_epoch=int((self.risetime-datetime(1970,1,1)).total_seconds())
        sat_lon, sat_lat, alt = self.orb.get_lonlatalt(self.risetime)

        ovpass = ET.SubElement(root, "pass")
        ovpass.set("satellite", self.satellite.name)
        ovpass.set("aos", self.risetime.strftime("%Y%m%d%H%M%S"))
        ovpass.set("los", self.falltime.strftime("%Y%m%d%H%M%S"))
        ovpass.set("orbit", "{:d}".format(orbit))
        ovpass.set("max-elevation", "{:.3f}".format(max_elevation))
        ovpass.set("asimuth-at-max-elevation", "{:.3f}".format(asimuth_at_max_elevation))
        ovpass.set("asimuth-at-aos", "{:.3f}".format(asimuth_at_aos))
        ovpass.set("pass-direction", pass_direction)
        ovpass.set("satellite-lon-at-aos", "{:.3f}".format(sat_lon))
        ovpass.set("satellite-lat-at-aos", "{:.3f}".format(sat_lat))
        ovpass.set("tle-epoch", self.orb.orbit_elements.epoch.astype(datetime).strftime("%Y%m%d%H%M%S.%f"))
        if self.fig:
            ovpass.set("figure", self.fig)

        return True

    def print_meos(self, coords, line_no):
        """No. Date    Satellite  Orbit Max EL  AOS      Ovlp  LOS      Durtn  Az(AOS/MAX)."""
        asimuth_at_max_elevation, max_elevation = self.orb.get_observer_look(self.uptime, *coords)
        pass_direction = self.pass_direction().capitalize()[:1]
        # anl = self.orb.get_lonlatalt(self.orb.get_last_an_time(self.risetime))[0] % 360
        asimuth_at_aos, aos_elevation = self.orb.get_observer_look(self.risetime, *coords)
        orbit = self.orb.get_orbit_number(self.risetime)
        aos_epoch = int((self.risetime - datetime(1970, 1, 1)).total_seconds())
        sat_lon, sat_lat, alt = self.orb.get_lonlatalt(self.risetime)

        dur_secs = (self.falltime - self.risetime).seconds
        dur_hours, dur_reminder = divmod(dur_secs, 3600)
        dur_minutes, dur_seconds = divmod(dur_reminder, 60)
        duration = "{:0>2}:{:0>2}".format(dur_minutes, dur_seconds)

        pass_key = hashlib.md5(("{:s}|{:d}|{:d}|{:.3f}|{:.3f}".  # noqa : md5 is insecure, but not sensitive here.
                                format(SATELLITE_MEOS_TRANSLATION.get(self.satellite.name.upper(),
                                                                      self.satellite.name.upper()),
                                       int(orbit),
                                       aos_epoch,
                                       sat_lon,
                                       sat_lat)).encode("utf-8")).hexdigest()

        line_list = [" {line_no:>2}",
                     "{date}",
                     "{satellite:<10}",
                     "{orbit:>5}",
                     "{elevation:>6.3f} ",
                     "{risetime}",
                     "{overlap:<5s}",
                     "{falltime}",
                     "{duration}",
                     "{asimuth_at_aos:>5.1f}",
                     "{asimuth_at_max:>5.1f}",
                     "-- Undefined(Scheduling not done {aos_epoch} )",
                     "{passkey}",
                     "{pass_direction}"
                     ]

        line = " ".join(line_list).format(
            # line_no=line_no,
            line_no=1,
            date=self.risetime.strftime("%Y%m%d"),
            satellite=SATELLITE_MEOS_TRANSLATION.get(self.satellite.name.upper(),
                                                     self.satellite.name.upper()),
            orbit=orbit,
            elevation=max_elevation,
            risetime=self.risetime.strftime("%H:%M:%S"),
            overlap="n/a",
            falltime=self.falltime.strftime("%H:%M:%S"),
            duration=duration,
            asimuth_at_aos=asimuth_at_aos,
            asimuth_at_max=asimuth_at_max_elevation,
            aos_epoch=aos_epoch,
            passkey=pass_key,
            pass_direction=pass_direction)
        return line

    def print_vcs(self, coords):
        """Print a vcs/scisys/cgi schedule.

        Should look like this::

        # SCName          RevNum Risetime        Falltime        Elev Dura ANL   Rec Dir Man Ovl OvlSCName
        #      OvlRev OvlRisetime     OrigRisetime    OrigFalltime    OrigDuration
        # NOAA 19           24845 20131204 001450 20131204 003003 32.0 15.2 225.6 Y   Des N   N   none
        #      0 19580101 000000 20131204 001450 20131204 003003 15.2


        """
        max_elevation = self.orb.get_observer_look(self.uptime, *coords)[1]
        anl = self.orb.get_lonlatalt(self.orb.get_last_an_time(
            self.risetime))[0] % 360
        # anl = self.orb.get_observer_look(self.risetime, *coords)[0]
        if self.rec:
            rec = "Y"
        else:
            rec = "N"
        line_list = [
            "{satellite:<16}",
            "{orbit:>6}",
            "{risetime}",
            "{falltime}",
            "{elevation:>4.1f}",
            "{duration:>4.1f}",
            "{anl:>5.1f}",
            "{rec:<3}",
            "{direction}",
            "N   N   none                  0 19580101 000000",
            "{risetime}",
            "{falltime}",
            "{duration:>4.1f}",
        ]
        line = " ".join(line_list).format(
            satellite=self.satellite.name.upper(),
            orbit=self.orb.get_orbit_number(self.risetime),
            risetime=self.risetime.strftime("%Y%m%d %H%M%S"),
            falltime=self.falltime.strftime("%Y%m%d %H%M%S"),
            elevation=max_elevation,
            duration=(self.falltime - self.risetime).seconds / 60.0,
            anl=anl,
            rec=rec,
            direction=self.pass_direction().capitalize()[:3])
        return line


HOST = "ftp://is.sci.gsfc.nasa.gov/ancillary/ephemeris/schedule/%s/downlink/"


def get_aqua_terra_dumps(start_time,
                         end_time,
                         satorb,
                         sat,
                         dump_url=None):
    """Get the Terra and Aqua overpasses.

     We take into account the fact that when
    there are global dumps there is no direct broadcast.
    """
    # Get the list of aqua/terra dump info:
    dump_info_list = get_aqua_terra_dumpdata_from_ftp(sat, dump_url)

    dumps = []
    for elem in dump_info_list:
        if elem["los"] >= start_time and elem["aos"] <= end_time:
            uptime = elem["aos"] + (elem["los"] - elem["aos"]) / 2
            overpass = Pass(sat, elem["aos"], elem["los"],
                            orb=satorb, uptime=uptime, instrument="modis")
            overpass.station = elem["station"]
            overpass.max_elev = elem["elev"]
            dumps.append(overpass)

    return dumps


def get_aqua_terra_dumpdata_from_ftp(sat, dump_url):
    """Get the information on the internet on the actual global dumps of Terra and Aqua."""
    logger.info("Fetch %s dump info from internet", str(sat.name))
    if isinstance(dump_url, str):
        url = urlparse(dump_url % sat.name)
    else:
        url = urlparse(HOST % sat.name)
    logger.debug("Connect to ftp server")
    try:
        f = ftplib.FTP_TLS(url.netloc)
    except (socket.error, socket.gaierror) as e:
        logger.error("cannot reach to %s " % url.netloc + str(e))
        f = None

    if f is not None:
        try:
            if sys.version_info < (2, 7, 10):
                f.ssl_version = ssl.PROTOCOL_SSLv23
            f.login("anonymous", "guest")
            logger.debug("Logged in")
        except ftplib.error_perm:
            logger.error("cannot login anonymously")
            f.quit()
            f = None

    if f is not None:
        data = []
        try:
            f.prot_p()  # explicitly call for protected transfer
            f.dir(url.path, data.append)
        except socket.error as e:
            logger.error("Can't get any data: " + str(e))
            f.quit()
            f = None
        else:
            filenames = [line.split()[-1] for line in data]

    if f is None:
        logger.info("Can't access ftp server, using cached data")
        filenames = glob.glob(os.path.join(gettempdir(), "*.rpt"))

    filenames = [
        x for x in filenames if x.startswith("wotis.") and x.endswith(".rpt")
    ]
    dates = [
        datetime.strptime("".join(filename.split(".")[2:4]), "%Y%j%H%M%S")
        for filename in filenames
    ]
    filedates = dict(zip(dates, filenames))

    dumps = []

    for date in sorted(dates):
        lines = []
        if not os.path.exists(os.path.join(gettempdir(), filedates[date])):
            try:
                f.prot_p()  # explicitly call for protected transfer
                f.retrlines("RETR " + os.path.join(url.path, filedates[date]),
                            lines.append)
            except ftplib.error_perm:
                logger.info("Permission error (???) on ftp server, skipping.")
                continue
            with open(os.path.join(gettempdir(), filedates[date]), "w") as fd_:
                for line in lines:
                    fd_.write(line + "\n")

        else:
            with open(os.path.join(gettempdir(), filedates[date]), "r") as fd_:
                for line in fd_:
                    lines.append(line)

        # for line in lines[7::2]:
        #     if line.strip() == '':
        #         break
        #     station, aos, elev, los = line.split()[:4]
        #     aos = datetime.strptime(aos, "%Y:%j:%H:%M:%S")
        #     los = datetime.strptime(los, "%Y:%j:%H:%M:%S")
        #     if los >= start_time and aos <= end_time:
        #         uptime = aos + (los - aos) / 2
        #         overpass = Pass(sat, aos, los, orb=satorb, uptime=uptime, instrument="modis")
        #         overpass.station = station
        #         overpass.max_elev = elev
        #         dumps.append(overpass)

        for line in lines[7::2]:
            if line.strip() == "":
                break
            station, aos, elev, los = line.split()[:4]
            aos = datetime.strptime(aos, "%Y:%j:%H:%M:%S")
            los = datetime.strptime(los, "%Y:%j:%H:%M:%S")
            dumps.append({"station": station, "aos": aos, "los": los, "elev": elev})

    if f is not None:
        f.quit()
    return dumps


def get_next_passes(satellites,
                    utctime,
                    forward,
                    coords,
                    tle_file=None,
                    aqua_terra_dumps=None,
                    min_pass=MIN_PASS,
                    local_horizon=0):
    """Get the next passes for *satellites*.

    Get the next passes for *satellites* , starting at *utctime*, for a
    duration of *forward* hours, with observer at *coords* ie lon (°E), lat
    (°N), altitude (km). Uses *tle_file* if provided, downloads from celestrack
    otherwise.

    Metop-A, Terra and Aqua need special treatment due to downlink restrictions.
    """
    passes = {}

    if tle_file is None and "TLES" not in os.environ:
        fp_, tle_file = mkstemp(prefix="tle", dir=gettempdir())
        os.close(fp_)
        logger.info("Fetch tle info from internet")
        tlefile.fetch(tle_file)

    if not os.path.exists(tle_file) and "TLES" not in os.environ:
        logger.info("Fetch tle info from internet")
        tlefile.fetch(tle_file)

    for sat in satellites:
        if not hasattr(sat, "name"):
            sat = Satellite(sat, 0, 0)

        satorb = orbital.Orbital(sat.name, tle_file=tle_file)
        passlist = satorb.get_next_passes(utctime,
                                          forward,
                                          *coords,
                                          horizon=local_horizon,
                                          )

        if sat.name.lower() == "metop-a":
            # Take care of metop-a special case
            passes["metop-a"] = get_metopa_passes(sat, passlist, satorb)
        elif sat.name.lower() in ["aqua", "terra"] and aqua_terra_dumps:
            # Take care of aqua (dumps in svalbard and poker flat)
            # Get the Terra/Aqua passes and fill the passes dict:
            get_terra_aqua_passes(passes, utctime, forward, sat, passlist, satorb, aqua_terra_dumps)
        else:
            if sat.name.upper() in VIIRS_PLATFORM_NAMES:
                instrument = "viirs"
            elif sat.name.lower().startswith("metop") or sat.name.lower().startswith("noaa"):
                instrument = "avhrr"
            elif sat.name.lower() in ["aqua", "terra"]:  # when aqua_terra_dumps=False
                instrument = "modis"
            elif sat.name.upper() in MERSI_PLATFORM_NAMES:
                instrument = "mersi"
            elif sat.name.upper() in MERSI2_PLATFORM_NAMES:
                instrument = "mersi-2"
            elif sat.name.upper() in MERSI3_PLATFORM_NAMES:
                instrument = "mersi-3"
            else:
                instrument = "unknown"

            passes[sat.name] = [
                Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument=instrument)
                for rtime, ftime, uptime in passlist
                if ftime - rtime > timedelta(minutes=min_pass)
            ]

    return set(fctools_reduce(operator.concat, list(passes.values())))


def get_metopa_passes(sat, passlist, satorb):
    """Get the Metop-A passes, taking care that Metop-A doesn't transmit to ground everywhere."""
    metop_passes = [
        Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument="avhrr")
        for rtime, ftime, uptime in passlist if rtime < ftime
    ]

    passes = []
    for overpass in metop_passes:
        if overpass.pass_direction() == "descending":
            new_rise = overpass.slsearch(60)
            if new_rise is not None and new_rise < overpass.falltime:
                overpass.risetime = new_rise
                # overpass has a boundary property, and it is not really needed here anyways!
                # overpass.boundary = SwathBoundary(overpass)
                if overpass.seconds() > MIN_PASS * 60:
                    passes.append(overpass)

    return passes


def get_terra_aqua_passes(passes, utctime, forward, sat, passlist, satorb, aqua_terra_dumps):
    """Get the Terra/Aqua passes.

    We take care that Terra and Aqua do not have direct broadcast when there are global dumps.

    Args:
        passes: The dictionary of satellite passes which is being built

        utctime: The start time (datetime object)

        forward: The number of hours ahead for which we will get the coming passes

        sat: The Satellite platform considered

        passlist: List of Pass objects

        satorb: Orbital instance for the actual satellite and tles considered

        aqua_terra_dumps: True or False or the actual URL to get info on Terra/Aqua
           dumps.  If True, the default URL will be used. If False or None, no dump
           info will be considered.

    """
    instrument = "modis"

    wpcoords = (-75.457222, 37.938611, 0)
    passlist_wp = satorb.get_next_passes(
        utctime - timedelta(minutes=30), forward + 1, *wpcoords)
    wp_passes = [
        Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument=instrument)
        for rtime, ftime, uptime in passlist_wp if rtime < ftime
    ]

    svcoords = (15.399, 78.228, 0)
    passlist_sv = satorb.get_next_passes(
        utctime - timedelta(minutes=30), forward + 1, *svcoords)
    sv_passes = [
        Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument=instrument)
        for rtime, ftime, uptime in passlist_sv if rtime < ftime
    ]
    pfcoords = (-147.43, 65.12, 0.51)
    passlist_pf = satorb.get_next_passes(
        utctime - timedelta(minutes=30), forward + 1, *pfcoords)
    pf_passes = [
        Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument=instrument)
        for rtime, ftime, uptime in passlist_pf if rtime < ftime
    ]

    aqua_passes = [
        Pass(sat, rtime, ftime, orb=satorb, uptime=uptime, instrument=instrument)
        for rtime, ftime, uptime in passlist if rtime < ftime
    ]

    dumps = get_aqua_terra_dumps(utctime - timedelta(minutes=30),
                                 utctime + timedelta(hours=forward + 0.5),
                                 satorb, sat, aqua_terra_dumps)

    # remove the known dumps
    for dump in dumps:
        # print "*", dump.station, dump, dump.max_elev
        logger.debug("dump from ftp: " + str((dump.station, dump,
                                              dump.max_elev)))
        for i, sv_pass in enumerate(sv_passes):
            if sv_pass.overlaps(dump, timedelta(minutes=40)):
                sv_elevation = sv_pass.orb.get_observer_look(
                    sv_pass.uptime, *svcoords)[1]
                logger.debug("Computed " + str(("SG", sv_pass,
                                                sv_elevation)))
                del sv_passes[i]
        for i, pf_pass in enumerate(pf_passes):
            if pf_pass.overlaps(dump, timedelta(minutes=40)):
                pf_elevation = pf_pass.orb.get_observer_look(
                    pf_pass.uptime, *pfcoords)[1]
                logger.debug("Computed " + str(("PF", pf_pass,
                                                pf_elevation)))
                del pf_passes[i]
        for i, wp_pass in enumerate(wp_passes):
            if wp_pass.overlaps(dump, timedelta(minutes=40)):
                wp_elevation = wp_pass.orb.get_observer_look(
                    wp_pass.uptime, *wpcoords)[1]
                logger.debug("Computed " + str(("WP", wp_pass,
                                                wp_elevation)))
                del wp_passes[i]

    # sort out dump passes first
    # between sv an pf, we take the one with the highest elevation if
    # pf < 20°, pf otherwise
    # I think wp is also used if sv is the only other alternative
    used_pf = []
    for sv_pass in sv_passes:
        found_pass = False
        for pf_pass in pf_passes:
            if sv_pass.overlaps(pf_pass):
                found_pass = True
                used_pf.append(pf_pass)
                sv_elevation = sv_pass.orb.get_observer_look(
                    sv_pass.uptime, *svcoords)[1]
                pf_elevation = pf_pass.orb.get_observer_look(
                    pf_pass.uptime, *pfcoords)[1]
                if pf_elevation > 20:
                    dumps.append(pf_pass)
                elif sv_elevation > pf_elevation:
                    dumps.append(sv_pass)
                else:
                    dumps.append(pf_pass)
                break
        if not found_pass:
            dumps.append(sv_pass)

    for pf_pass in pf_passes:
        if pf_pass not in used_pf:
            dumps.append(pf_pass)

    passes[sat.name] = []
    for overpass in aqua_passes:
        add = True
        for dump_pass in dumps:
            if dump_pass.overlaps(overpass):
                if (dump_pass.uptime < overpass.uptime and
                        dump_pass.falltime > overpass.risetime):
                    logger.debug("adjusting " + str(overpass) +
                                 " to new risetime " +
                                 str(dump_pass.falltime))
                    overpass.risetime = dump_pass.falltime
                    overpass.boundary = SwathBoundary(overpass)
                elif (dump_pass.uptime >= overpass.uptime and
                      dump_pass.risetime < overpass.falltime):
                    logger.debug("adjusting " + str(overpass) +
                                 " to new falltime " +
                                 str(dump_pass.risetime))
                    overpass.falltime = dump_pass.risetime
                    overpass.boundary = SwathBoundary(overpass)
                if overpass.falltime <= overpass.risetime:
                    add = False
                    logger.debug("skipping " + str(overpass))
        if add and overpass.seconds() > MIN_PASS * 60:
            passes[sat.name].append(overpass)

    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2014 - 2025 PyTroll Community

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Package file."""


# shortest allowed pass in minutes
MIN_PASS = 4

# DRL still use the name JPSS-1 (etc) instead of NOAA-20 in the TLEs:
JPSS_TLE_NAMES = {"NOAA-20": "JPSS-1",
                  "NOAA-21": "JPSS-2",
                  "NOAA-22": "JPSS-3"}

NUMBER_OF_FOVS = {
    "avhrr": 2048,
    "mhs": 90,
    "amsua": 30,
    "mwhs2": 98,
    "atms": 96,
    "ascat": 42,
    "viirs": 6400,
    "mwhs-2": 98
}

SATELLITE_NAMES = {"npp": "Suomi NPP",
                   "noaa19": "NOAA 19",
                   "noaa18": "NOAA 18",
                   "noaa15": "NOAA 15",
                   "aqua": "Aqua",
                   "terra": "Terra",
                   "metopc": "Metop-C",
                   "metopb": "Metop-B",
                   "metopa": "Metop-A",
                   "noaa20": "NOAA-20",
                   "noaa21": "NOAA-21",
                   "noaa22": "NOAA-22",
                   "fengyun3d": "FY-3D",
                   "fengyun3c": "FY-3C",
                   "fengyun3e": "FY-3E",
                   "fengyun3f": "FY-3F"
                   }

INSTRUMENT = {"Suomi NPP": "viirs",
              "NOAA-20": "viirs",
              "NOAA-21": "viirs",
              "NOAA-22": "viirs",
              "Aqua": "modis",
              "Terra": "modis",
              "NOAA 19": "avhrr",
              "NOAA 18": "avhrr",
              "NOAA 15": "avhrr",
              "Metop-A": "avhrr",
              "Metop-B": "avhrr",
              "Metop-C": "avhrr",
              "FY-3D": "avhrr",
              "FY-3C": "avhrr"}

VIIRS_PLATFORM_NAMES = ["SUOMI NPP", "SNPP",
                        "NOAA-20", "NOAA 20", "NOAA-21", "NOAA 21"]

MERSI_PLATFORM_NAMES = ["FENGYUN 3C", "FENGYUN-3C", "FY-3C"]
MERSI2_PLATFORM_NAMES = ["FENGYUN 3D", "FENGYUN-3D", "FY-3D",
                         "FENGYUN 3E", "FENGYUN-3E", "FY-3E"]
MERSI3_PLATFORM_NAMES = ["FENGYUN 3F", "FENGYUN-3F", "FY-3F"]

SATELLITE_MEOS_TRANSLATION = {"NOAA 19": "NOAA_19",
                              "NOAA 18": "NOAA_18",
                              "NOAA 15": "NOAA_15",
                              "METOP-A": "M02",
                              "METOP-B": "M01",
                              "FENGYUN 3A": "FENGYUN-3A",
                              "FENGYUN 3B": "FENGYUN-3B",
                              "FENGYUN 3C": "FENGYUN-3C",
                              "SUOMI NPP": "NPP"}



# Copyright (c) 2013 - 2024 Pytroll community

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Alexander Maul <alexander.maul@dwd.de>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module and script for pass scheduling."""
import argparse
import logging
import logging.handlers
import os
from datetime import datetime, timedelta
from pprint import pformat
from urllib.parse import urlparse

import numpy as np
from pyorbital import astronomy
from pyresample import parse_area_file

from trollsched import MIN_PASS, utils
from trollsched.combine import get_combined_sched
from trollsched.graph import Graph
from trollsched.satpass import SimplePass, get_next_passes
from trollsched.spherical import get_twilight_poly
from trollsched.writers import generate_meos_file, generate_metno_xml_file, generate_sch_file, generate_xml_file

logger = logging.getLogger(__name__)


class Station:
    """docstring for Station."""

    def __init__(self, station_id, name, longitude, latitude, altitude, area, satellites, area_file=None,
                 min_pass=MIN_PASS, local_horizon=0):
        """Initialize the station."""
        self.id = station_id
        self.name = name
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude
        self.area = area
        self.satellites = satellites

        if area_file is not None:
            try:
                self.area = parse_area_file(area_file, area)[0]
            except TypeError:
                pass
        self.min_pass = min_pass
        self.local_horizon = local_horizon

    @property
    def coords(self):
        """Get the coordinates lon, lat, alt."""
        return self.longitude, self.latitude, self.altitude

    def single_station(self, sched, start_time, tle_file):
        """Calculate passes, graph, and schedule for one station."""
        logger.debug("station: %s coords: %s area: %s scores: %s",
                     self.id, self.coords, self.area.area_id, self.satellites)

        opts = sched.opts
        pattern = sched.patterns
        pattern_args = {
            "station": self.id,
            "output_dir": opts.output_dir,
            "date": start_time.strftime("%Y%m%d"),
            "time": start_time.strftime("%H%M%S")
        }
        if opts.xml:
            pattern_args["mode"] = "request"
        elif opts.report:
            pattern_args["mode"] = "report"

        allpasses = self.get_next_passes(opts, sched, start_time, tle_file)

        area_boundary = self.area.boundary(8)
        self.area.poly = area_boundary.contour_poly

        if opts.plot:
            logger.info("Saving plots to %s", build_filename(
                "dir_plots", pattern, pattern_args))
            from threading import Thread
            image_saver = Thread(
                target=save_passes,
                args=(allpasses,
                      self.area.poly,
                      build_filename(
                          "dir_plots", pattern, pattern_args),
                      sched.plot_parameters,
                      sched.plot_title
                      )
            )
            image_saver.start()

        if opts.avoid is not None:
            avoid_list = get_passes_from_xml_file(opts.avoid)
        else:
            avoid_list = None

        logger.info("computing best schedule for area %s" % self.area.area_id)
        schedule, (graph, labels) = get_best_sched(allpasses,
                                                   self.area,
                                                   timedelta(seconds=opts.delay),
                                                   avoid_list)

        logger.debug(pformat(schedule))
        for opass in schedule:
            opass.rec = True
        logger.info("generating file")

        if opts.scisys:
            generate_sch_file(build_filename("file_sci", pattern,
                                             pattern_args), allpasses, self.coords)

        if opts.meos:
            generate_meos_file(build_filename("file_meos", pattern, pattern_args), allpasses,
                               self.coords, start_time + timedelta(hours=sched.start), True)  # Ie report mode

        if opts.plot:
            logger.info("Waiting for images to be saved...")
            image_saver.join()
            logger.info("Done!")

        if opts.metno_xml:
            generate_metno_xml_file(build_filename("file_metno_xml_all", pattern, pattern_args), allpasses,
                                    self.coords, start_time + timedelta(hours=sched.start),
                                    start_time + timedelta(hours=sched.forward), self.id, sched.center_id,
                                    report_mode=True)
            generate_metno_xml_file(build_filename("file_metno_xml", pattern, pattern_args), schedule,
                                    self.coords, start_time + timedelta(hours=sched.start),
                                    start_time + timedelta(hours=sched.forward), self.id, sched.center_id,
                                    report_mode=True)

        if opts.xml or opts.report:
            url = urlparse(opts.output_url or opts.output_dir)
            if opts.xml or opts.report:
                # Always create xml-file in request-mode
                pattern_args["mode"] = "request"
                xmlfile = generate_xml_file(allpasses,
                                            start_time + timedelta(hours=sched.start),
                                            start_time + timedelta(hours=sched.forward),
                                            build_filename(
                                                "file_xml", pattern, pattern_args),
                                            self.id,
                                            sched.center_id,
                                            report_mode=False
                                            )
                logger.info("Generated " + str(xmlfile))
                send_file(url, xmlfile)
            if opts.report:
                """'If report-mode was set"""
                pattern_args["mode"] = "report"
                xmlfile = generate_xml_file(allpasses,
                                            start_time + timedelta(hours=sched.start),
                                            start_time + timedelta(hours=sched.forward),
                                            build_filename(
                                                "file_xml", pattern, pattern_args),
                                            self.id,
                                            sched.center_id,
                                            True
                                            )
                logger.info("Generated " + str(xmlfile))

        if opts.graph or opts.comb:
            graph.save(build_filename("file_graph", pattern, pattern_args))
            graph.export(
                labels=[str(label) for label in labels],
                filename=build_filename("file_graph", pattern, pattern_args) + ".gv"
            )
        if opts.comb:
            import pickle
            ph = open(os.path.join(build_filename("dir_output", pattern,
                                                  pattern_args), "allpasses.%s.pkl" % self.id), "wb")
            pickle.dump(allpasses, ph)
            ph.close()

        return graph, allpasses

    def get_next_passes(self, opts, sched, start_time, tle_file):
        """Get the next passes."""
        logger.info("Computing next satellite passes")
        allpasses = get_next_passes(self.satellites, start_time,
                                    sched.forward,
                                    self.coords, tle_file,
                                    aqua_terra_dumps=(sched.dump_url or True
                                                      if opts.no_aqua_terra_dump
                                                      else None),
                                    min_pass=self.min_pass,
                                    local_horizon=self.local_horizon
                                    )
        logger.info("Computation of next overpasses done")
        logger.debug(str(sorted(allpasses, key=lambda x: x.risetime)))
        return allpasses



class Scheduler:
    """docstring for Scheduler."""

    def __init__(self, stations, min_pass, forward, start, dump_url, patterns, center_id, plot_parameters, plot_title):
        """Initialize the scheduler."""
        self.stations = stations
        self.min_pass = min_pass
        self.forward = forward
        self.start = start
        self.dump_url = dump_url
        self.patterns = patterns
        self.center_id = center_id
        self.plot_parameters = plot_parameters
        self.plot_title = plot_title
        self.opts = None


def conflicting_passes(allpasses, delay=None):
    """Get the passes in groups of conflicting passes."""
    if delay is None:
        delay = timedelta(seconds=0)

    passes = sorted(allpasses, key=lambda x: x.risetime)

    overpass = passes[0]
    last_time = overpass.falltime
    group = [overpass]
    groups = []
    for overpass in passes[1:]:
        if overpass.risetime - delay < last_time:
            group.append(overpass)
            if last_time < overpass.falltime:
                last_time = overpass.falltime

        else:
            groups.append(group)
            group = [overpass]
            last_time = overpass.falltime

    groups.append(group)
    return groups


def get_non_conflicting_groups(passes, delay=None):
    """Get the different non-conflicting solutions in a group of conflicting passes."""
    # Uses graphs and maximal clique finding with the Bron-Kerbosch algorithm.
    if delay is None:
        delay = timedelta(seconds=0)

    order = len(passes)

    if order == 1:
        return [passes]

    graph = Graph(order)

    for i, overpass in enumerate(sorted(passes, key=lambda x: x.risetime)):
        for j in range(i + 1, order):
            if not overpass.overlaps(passes[j], delay):
                graph.add_edge(i, j)

    groups = []
    for res in graph.bron_kerbosch(set(), set(graph.vertices), set()):
        grp = []
        for vertex in res:
            grp.append(passes[vertex])
        groups.append(sorted(grp))

    return groups


def fermia(t):
    """Return the Fermi value a."""
    a = 0.25
    b = a / 4
    k = b * np.log(1 / 3.0) + a
    sh = k - 0.25
    return 0.5 / (np.exp(((t + sh) - a) / b) + 1) + 0.5


def fermib(t):
    """Return the Fermi value b."""
    a = 0.25
    b = a / 4
    return 1 / (np.exp((t - a) / b) + 1)


combination: dict[SimplePass, SimplePass] = {}


def combine(p1, p2, area_of_interest):
    """Combine passes together."""
    try:
        return combination[p1, p2]
    except KeyError:
        pass

    area = area_of_interest.poly.area()

    def pscore(poly, coeff=1):
        if poly is None:
            return 0
        else:
            return poly.area() * coeff

    twi1 = get_twilight_poly(p1.uptime)
    twi2 = get_twilight_poly(p2.uptime)

    ip1, sip1 = p1.score.get(area_of_interest, (None, None))
    if sip1 is None:
        ip1 = p1.boundary.contour_poly.intersection(area_of_interest.poly)
        # FIXME: ip1 or ip2 could be None if the pass is entirely inside the
        # area (or vice versa)
        if ip1 is None:
            return 0

        ip1d = ip1.intersection(twi1)
        if ip1d is None:
            lon, lat = np.rad2deg(ip1.vertices[0, :])
            theta = astronomy.cos_zen(p1.uptime,
                                      lon, lat)
            if np.sign(theta) > 0:
                ip1d = ip1
                ip1n = None
            else:
                ip1n = ip1
        else:
            twi1.invert()
            ip1n = ip1.intersection(twi1)
            twi1.invert()

        ns1 = pscore(ip1n, p1.satellite.score.night / area)
        ds1 = pscore(ip1d, p1.satellite.score.day / area)
        sip1 = ns1 + ds1
        p1.score[area_of_interest] = (ip1, sip1)

    ip2, sip2 = p2.score.get(area_of_interest, (None, None))
    if sip2 is None:
        ip2 = p2.boundary.contour_poly.intersection(area_of_interest.poly)
        if ip2 is None:
            return 0

        ip2d = ip2.intersection(twi2)
        if ip2d is None:
            lon, lat = np.rad2deg(ip2.vertices[0, :])
            theta = astronomy.cos_zen(p2.uptime,
                                      lon, lat)
            if np.sign(theta) > 0:
                ip2d = ip2
                ip2n = None
            else:
                ip2n = ip2
        else:
            twi2.invert()
            ip2n = ip2.intersection(twi2)
            twi2.invert()

        ns2 = pscore(ip2n, p2.satellite.score.night / area)
        ds2 = pscore(ip2d, p2.satellite.score.day / area)
        sip2 = ns2 + ds2
        p2.score[area_of_interest] = (ip2, sip2)

    ip1p2 = ip1.intersection(ip2)

    if ip1p2 is None:
        sip1p2 = 0
    else:
        ip1p2da = ip1p2.intersection(twi1)
        twi1.invert()
        ip1p2na = ip1p2.intersection(twi1)
        twi1.invert()

        ip1p2db = ip1p2.intersection(twi2)
        twi2.invert()
        ip1p2nb = ip1p2.intersection(twi2)
        twi2.invert()

        ns12a = pscore(ip1p2na, p1.satellite.score.night / area)
        ds12a = pscore(ip1p2da, p1.satellite.score.day / area)
        ns12b = pscore(ip1p2nb, p2.satellite.score.night / area)
        ds12b = pscore(ip1p2db, p2.satellite.score.day / area)

        sip1p2a = ns12a + ds12a
        sip1p2b = ns12b + ds12b
        sip1p2 = (sip1p2a + sip1p2b) / 2.0

    if p2 > p1:
        tdiff = (p2.uptime - p1.uptime).seconds / 3600.
    else:
        tdiff = (p1.uptime - p2.uptime).seconds / 3600.

    res = fermia(tdiff) * (sip1 + sip2) - fermib(tdiff) * sip1p2
    combination[p1, p2] = res

    return res


def get_best_sched(overpasses, area_of_interest, delay, avoid_list=None):
    """Get the best schedule based on *area_of_interest*."""
    avoid_list = avoid_list or []
    passes = sorted(overpasses, key=lambda x: x.risetime)
    grs = conflicting_passes(passes, delay)
    logger.debug("conflicting %s", str(grs))
    ncgrs = [get_non_conflicting_groups(gr, delay) for gr in grs]
    logger.debug("non conflicting %s", str(ncgrs))
    n_vertices = len(passes)

    graph = Graph(n_vertices=n_vertices + 2)

    def add_arc(graph, p1, p2, hook=None):
        logger.debug(f"Adding arc between {p1} and {p2}...")
        if p1 in avoid_list or p2 in avoid_list:
            w = 0
            logger.debug("...0 because in the avoid_list!")
        else:
            w = combine(p1, p2, area_of_interest)
        logger.debug("...with weight " + str(w))

        graph.add_arc(passes.index(p1) + 1, passes.index(p2) + 1, w)
        if hook is not None:
            hook()

    prev = set()
    for ncgr in ncgrs:
        for pr in prev:
            foll = set(gr[0] for gr in ncgr)
            for f in foll:
                add_arc(graph, pr, f)

        prev = set(sorted(gr, key=lambda x: x.falltime)[-1] for gr in ncgr)
        for gr in ncgr:
            if len(gr) > 1:
                for p1, p2 in zip(gr[:-1], gr[1:]):
                    add_arc(graph, p1, p2)

    for pr in prev:
        graph.add_arc(passes.index(pr) + 1, n_vertices + 1)
    for first in ncgrs[0][0]:
        graph.add_arc(0, passes.index(first) + 1)

    dist, path = graph.dag_longest_path(0, n_vertices + 1)

    del dist
    return [passes[idx - 1] for idx in path[1:-1]], (graph, passes)


def argmax(iterable):
    """Find the index of the maximum of an iterable."""
    return max((x, i) for i, x in enumerate(iterable))[1]


def get_max(groups, fun):
    """Get the best group of *groups* using the score function *fun*."""
    scores = []
    for grp in groups:
        scores.append(sum([fun(p) for p in grp]))
    return groups[argmax(scores)]


def save_passes(allpasses, poly, output_dir, plot_parameters=None, plot_title=None):
    """Save overpass plots to png and store in directory *output_dir*."""
    from trollsched.drawing import save_fig
    for overpass in allpasses:
        save_fig(overpass, poly=poly, directory=output_dir, plot_parameters=plot_parameters, plot_title=plot_title)
    logger.info("All plots saved!")


def get_passes_from_xml_file(filenames):
    """Read passes from aquisition xml files."""
    import defusedxml.ElementTree as ET
    pass_list = []
    for filename in filenames:
        tree = ET.parse(filename)
        root = tree.getroot()
        for overpass in root.iter("pass"):
            start_time = datetime.strptime(
                overpass.attrib["start-time"], "%Y-%m-%d-%H:%M:%S")
            end_time = datetime.strptime(
                overpass.attrib["end-time"], "%Y-%m-%d-%H:%M:%S")
            pass_list.append(SimplePass(
                overpass.attrib["satellite"], start_time, end_time))
    return pass_list


def build_filename(pattern_name, pattern_dict, kwargs):
    """Build absolute path from pattern dictionary."""
    for k in pattern_dict.keys():
        for v in pattern_dict.values():
            if "{" + k + "}" in v:
                kwargs[k] = pattern_dict[k].format(**kwargs)

    return pattern_dict[pattern_name].format(**kwargs)


def send_file(url, file):
    """Send a file through ftp."""
    pathname, filename = os.path.split(file)
    del pathname
    if url.scheme in ["file", "", b""]:
        pass
    elif url.scheme in ["ftp", b"ftp"]:
        import ftplib
        session = ftplib.FTP(url.hostname, url.username, url.password)
        with open(file, "rb") as xfile:
            session.storbinary("STOR " + str(filename), xfile)
        session.quit()
    else:
        logger.error("Cannot save to %s, but file is there:",
                     (str(url.scheme), str(file)))


def combined_stations(scheduler, start_time, graph, allpasses):
    """The works around the combination of schedules for two or more stations."""
    logger.info("Generating coordinated schedules ...")

    def collect_labels(newpasses, stats):
        """Collect labels, each with one pass per station."""
        # TODO: is there a simpler way?
        clabels = []
        npasses = {s: set() for s in stats}
        for npass in newpasses:
            cl = []
            for i, s in zip(range(len(stats)), stats):
                if npass[i][0] is None:
                    cl.append("---")
                else:
                    npasses[s].add(npass[i][0])
                    if npass[i][0].rec:
                        cl.append("+ " + str(npass[i][0]))
                    else:
                        cl.append("  " + str(npass[i][0]))
            clabels.append("\\n".join(cl))
        return clabels

    pattern_args = {
        "output_dir": scheduler.opts.output_dir,
        "date": start_time.strftime("%Y%m%d"),
        "time": start_time.strftime("%H%M%S")
    }
    if scheduler.opts.xml:
        pattern_args["mode"] = "request"
    elif scheduler.opts.report:
        pattern_args["mode"] = "report"

    passes = {}
    # reset flag "rec" for all passes.
    try:
        for s, ap in allpasses.items():
            passes[s] = list(ap)
            for p in passes[s]:
                p.rec = False
    except Exception:
        logger.exception("Failed to reset 'rec' for s:%s  ap:%s  passes[s]:%s  p:%s",
                         s, ap, passes[s], p)
        raise

    stats, schedule, (newgraph, newpasses) = get_combined_sched(graph, passes)

    for opass in schedule:
        for _i, ipass in zip(range(len(opass)), opass):
            if ipass[0] is None:
                continue
            ipass[0].rec = True

    logger.info("generating files")

    if scheduler.opts.graph:
        # save graph as npz file.
        pattern_args["station"] = "comb"
        newgraph.save(build_filename("file_graph", scheduler.patterns, pattern_args))
        # Collect labels, each with one pass per station.
        clabels = collect_labels(newpasses, stats)
        # save graph as gv file for "dot"-plot
        newgraph.export(labels=[str(label) for label in clabels],
                        filename=build_filename("file_graph", scheduler.patterns, pattern_args) + ".gv")

    for station_id in passes.keys():
        pattern_args["station"] = station_id + "-comb"
        logger.info("Create schedule file(s) for %s", station_id)
        if scheduler.opts.scisys:
            generate_sch_file(build_filename("file_sci", scheduler.patterns, pattern_args),
                              passes[station_id],
                              [s.coords for s in scheduler.stations if s.id == station_id][0])
        if scheduler.opts.xml or scheduler.opts.report:
            pattern_args["mode"] = "request"
            xmlfile = generate_xml_file(passes[station_id],
                                        start_time + timedelta(hours=scheduler.start),
                                        start_time + timedelta(hours=scheduler.forward),
                                        build_filename(
                                            "file_xml", scheduler.patterns, pattern_args),
                                        station_id,
                                        scheduler.center_id,
                                        False)
            logger.info("Generated " + str(xmlfile))
            url = urlparse(scheduler.opts.output_url or scheduler.opts.output_dir)
            send_file(url, xmlfile)
        if scheduler.opts.report:
            pattern_args["mode"] = "report"
            xmlfile = generate_xml_file(passes[station_id],
                                        start_time + timedelta(hours=scheduler.start),
                                        start_time + timedelta(hours=scheduler.forward),
                                        build_filename(
                                            "file_xml", scheduler.patterns, pattern_args),
                                        # scheduler.stations[station_id].name,
                                        station_id,
                                        scheduler.center_id,
                                        True)
            logger.info("Generated " + str(xmlfile))

        if scheduler.opts.meos:
            meosfile = generate_meos_file(build_filename("file_meos", scheduler.patterns, pattern_args),
                                          passes[station_id],
                                          # station_meta[station]['coords'],
                                          [s.coords for s in scheduler.stations if s.id == station_id][0],
                                          start_time + timedelta(hours=scheduler.start),
                                          False)  # Ie only print schedule passes
            logger.info("Generated " + str(meosfile))
        if scheduler.opts.metno_xml:
            metno_xmlfile = generate_metno_xml_file(build_filename("file_metno_xml", scheduler.patterns, pattern_args),
                                                    passes[station_id],
                                                    # station_meta[station]['coords'],
                                                    [s.coords for s in scheduler.stations if s.id == station_id][0],
                                                    start_time + timedelta(hours=scheduler.start),
                                                    start_time + timedelta(hours=scheduler.forward),
                                                    station_id, scheduler.center_id, False)
            logger.info("Generated " + str(metno_xmlfile))

    logger.info("Finished coordinated schedules.")


def run(args=None):
    """The schedule command."""
    global logger

    opts = parse_args(args)

    if opts.config:
        # read_config() returns:
        #     [(coords, station, area, scores)], forward, start, {pattern}
        # station_list, forward, start, pattern = utils.read_config(opts.config)
        scheduler = utils.read_config(opts.config)

    if opts.output_dir:
        scheduler.patterns["dir_output"] = opts.output_dir
    else:
        scheduler.patterns.setdefault("dir_output", os.path.curdir)

    setup_logging(opts)

    tle_file = opts.tle
    if opts.start_time:
        start_time = opts.start_time
    else:
        start_time = datetime.utcnow()

    allpasses = {}
    graph = {}

    logger.debug("start: %s forward: %s" % (scheduler.start, scheduler.forward))

    pattern_args = {
        "output_dir": opts.output_dir,
        "date": start_time.strftime("%Y%m%d"),
        "time": start_time.strftime("%H%M%S")
    }
    dir_output = build_filename("dir_output", scheduler.patterns, pattern_args)
    if not os.path.exists(dir_output):
        logger.debug("Create output dir " + dir_output)
        os.makedirs(dir_output)

    if len(scheduler.stations) > 1:
        opts.comb = True
        import pickle
        ph = open(os.path.join(dir_output, "opts.pkl"), "wb")
        pickle.dump(opts, ph)
        ph.close()
    else:
        opts.comb = False

    scheduler.opts = opts

    # single- or multi-processing?
    if not opts.multiproc or len(scheduler.stations) == 1:
        # sequential processing all stations' single schedule.
        for station in scheduler.stations:
            graph[station.id], allpasses[station.id] = station.single_station(scheduler, start_time, tle_file)
    else:
        # processing the stations' single schedules with multiprocessing.
        process_single = {}
        statlst_ordered = []
        # first round through the stations, forking sub-processes to do the
        # "single station calculations" in parallel.
        # the pickling of passes and graphs is done inside single_station().
        for station in scheduler.stations:
            statlst_ordered.append(station.id)
            from multiprocessing import Process
            process_single[station.id] = Process(
                target=station.single_station,
                args=(scheduler, start_time, tle_file))
            process_single[station.id].start()
        # second round through the stations, collecting the sub-processes and
        # their results.
        for station_id in statlst_ordered:
            process_single[station_id].join()
            pattern_args["station"] = station_id
            # load graph for station
            graph[station_id] = Graph()
            graph[station_id].load(build_filename(
                "file_graph", scheduler.patterns, pattern_args) + ".npz")
            # load pickled passes for station
            ph = open(os.path.join(
                dir_output, "allpasses.%s.pkl" % station_id), "rb")
            allpasses[station_id] = pickle.load(ph)
            ph.close()

    if opts.comb:
        combined_stations(scheduler, start_time, graph, allpasses)


def setup_logging(opts):
    """Set up the logging."""
    global logger
    if opts.log:
        previous = os.path.exists(opts.log)
        handler = logging.handlers.RotatingFileHandler(opts.log, backupCount=7)
        if previous:
            handler.doRollover()
    else:
        handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("[%(levelname)s: %(asctime)s :"
                                           " %(name)s] %(message)s",
                                           "%Y-%m-%d %H:%M:%S"))
    if opts.verbose:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.INFO
    handler.setLevel(loglevel)
    logging.getLogger("").setLevel(loglevel)
    logging.getLogger("").addHandler(handler)
    if opts.mail:
        mhandler = logging.handlers.SMTPHandler("localhost",
                                                "pytroll-schedule@pytroll.org",
                                                opts.mail,
                                                "Scheduler")
        mhandler.setLevel(logging.WARNING)
        logging.getLogger("").addHandler(mhandler)
    logger = logging.getLogger("trollsched")


def parse_args(args=None):
    """Parse arguments from the command line."""
    parser = argparse.ArgumentParser()
    # general arguments
    parser.add_argument("-c", "--config", required=True, default=None,
                        help="configuration file to use")
    parser.add_argument("-t", "--tle", default=None,
                        help="tle file to use")
    parser.add_argument("-l", "--log", default=None,
                        help="File to log to (defaults to stdout)")
    parser.add_argument("-m", "--mail", nargs="*", default=None,
                        help="mail address(es) to send error messages to.")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="print debug messages too")
    # argument group: coordinates and times
    group_postim = parser.add_argument_group(title="start-parameter",
                                             description="(or set values in the configuration file)")
    group_postim.add_argument("--lat", type=float,
                              help="Latitude, degrees north")
    group_postim.add_argument("--lon", type=float,
                              help="Longitude, degrees east")
    group_postim.add_argument("--alt", type=float,
                              help="Altitude, km")
    group_postim.add_argument("-f", "--forward", type=float,
                              help="time ahead to compute the schedule")
    group_postim.add_argument("-s", "--start-time", type=datetime.fromisoformat,
                              help="start time of the schedule to compute")
    group_postim.add_argument("-d", "--delay", default=60, type=float,
                              help="delay (in seconds) needed between two " +
                                   "consecutive passes (60 seconds by default)")
    # argument group: special behaviour
    group_spec = parser.add_argument_group(title="special",
                                           description="(additional parameter changing behaviour)")
    group_spec.add_argument("-a", "--avoid",
                            default=[],
                            nargs="*",
                            help="xml request file(s) with passes to avoid")
    group_spec.add_argument("--no-aqua-terra-dump", action="store_false",
                            help="do not consider Aqua/Terra-dumps")
    group_spec.add_argument("--multiproc", action="store_true",
                            help="use multiple parallel processes")
    # argument group: output-related
    group_outp = parser.add_argument_group(title="output",
                                           description="(file pattern are taken from configuration file)")
    group_outp.add_argument("-o", "--output-dir", default=None,
                            help="where to put generated files")
    group_outp.add_argument("-u", "--output-url", default=None,
                            help="URL where to put generated schedule file(s)" +
                                 ", otherwise use output-dir")
    group_outp.add_argument("-x", "--xml", action="store_true",
                            help="generate an xml request file (schedule)"
                            )
    group_outp.add_argument("-r", "--report", action="store_true",
                            help="generate an xml report file (schedule)")
    group_outp.add_argument("--scisys", action="store_true",
                            help="generate a SCISYS schedule file")
    group_outp.add_argument("-p", "--plot", action="store_true",
                            help="generate plot images")
    group_outp.add_argument("-g", "--graph", action="store_true",
                            help="save graph info")
    group_outp.add_argument("--meos", action="store_true",
                            help="generate a MEOS schedule file")
    group_outp.add_argument("--metno-xml", action="store_true",
                            help="generate a METNO xml pass data file")
    opts = parser.parse_args(args)

    if (not opts.config) and (not (opts.lon or opts.lat or opts.alt)):
        parser.error("Coordinates must be provided in the absence of "
                     "configuration file.")

    if not (opts.xml or opts.scisys or opts.report or opts.metno_xml or opts.meos):
        parser.error("No output specified, use '--scisys', '-x/--xml', '-r/--report', '--meos', or '--metno-xml'")

    return opts


if __name__ == "__main__":
    try:
        run()
    except Exception:
        logger.exception("Something wrong happened!")
        raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2017, 2018, 2024 Alexander Maul
#
# Author(s):
#
#   Alexander Maul <alexander.maul@dwd.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions and config reading for the pytroll-scheduler."""
import logging
from collections.abc import Mapping

import yaml

from trollsched import schedule
from trollsched.pass_scheduling_utils import Satellite

logger = logging.getLogger("trollsched")


def read_yaml_file(file_name):
    """Read one or more files in to a single dict object."""
    if isinstance(file_name, str):
        file_name = [file_name]
    conf_dict = {}
    for file_obj in file_name:
        with open(file_obj) as fp:
            tmp_dict = yaml.safe_load(fp)
        conf_dict = recursive_dict_update(conf_dict, tmp_dict)
    return conf_dict


def recursive_dict_update(d, u):
    """Recursive dictionary update.

    Copied from:
        http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def read_config(filename):
    try:
        return read_config_yaml(filename)
    except yaml.parser.ParserError:
        logger.error("INI format for scheduler config is deprecated since v0.3.4, "
                     "please update your configuration to YAML.")
        raise


def read_config_yaml(filename):
    """Read the yaml file *filename* and create a scheduler."""
    cfg = read_yaml_file(filename)
    satellites = {sat_name: Satellite(sat_name, **sat_params)
                  for (sat_name, sat_params) in cfg["satellites"].items()}

    stations = {}
    for station_id, station in cfg["stations"].items():
        if isinstance(station["satellites"], dict):
            sat_list = []
            for (sat_name, sat_params) in station["satellites"].items():
                if sat_params is None:
                    sat_list.append(satellites[sat_name])
                else:
                    sat_list.append(Satellite(sat_name, **sat_params))
        else:
            sat_list = [satellites[sat_name] for sat_name in station["satellites"]]
        new_station = schedule.Station(station_id, **station)
        new_station.satellites = sat_list
        stations[station_id] = new_station

    pattern = {}
    for k, v in cfg["pattern"].items():
        pattern[k] = v

    sched_params = cfg["default"]
    plot_parameters = sched_params.get("plot_parameters", {})
    plot_title = sched_params.get("plot_title", None)

    scheduler = schedule.Scheduler(stations=[stations[st_id]
                                             for st_id in sched_params["station"]],
                                   min_pass=sched_params.get("min_pass", 4),
                                   forward=sched_params["forward"],
                                   start=sched_params["start"],
                                   dump_url=sched_params.get("dump_url"),
                                   patterns=pattern,
                                   center_id=sched_params.get("center_id", "unknown"),
                                   plot_parameters=plot_parameters,
                                   plot_title=plot_title)

    return scheduler
